(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports) :
	typeof define === 'function' && define.amd ? define(['exports'], factory) :
	(factory((global.combinatorsP = {})));
}(this, (function (exports) { 'use strict';

/**
 * Lift a value into a promise. This is equivalent to `Promise.resolve`.
 *
 * @example
 * const x = future(23);
 * const f = a => future(a + 1);
 *
 * @param {(Promise.<*>|*)} x The value to lift into a promise. This can
 * either be a value, or a promise that resolves to a value.
 * @returns {Promise.<*>} The value inside a promise.
 */
var future = function future(x) {
  return Promise.resolve(x);
};

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};



















var defineProperty = function (obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
};





















var slicedToArray = function () {
  function sliceIterator(arr, i) {
    var _arr = [];
    var _n = true;
    var _d = false;
    var _e = undefined;

    try {
      for (var _i = arr[Symbol.iterator](), _s; !(_n = (_s = _i.next()).done); _n = true) {
        _arr.push(_s.value);

        if (i && _arr.length === i) break;
      }
    } catch (err) {
      _d = true;
      _e = err;
    } finally {
      try {
        if (!_n && _i["return"]) _i["return"]();
      } finally {
        if (_d) throw _e;
      }
    }

    return _arr;
  }

  return function (arr, i) {
    if (Array.isArray(arr)) {
      return arr;
    } else if (Symbol.iterator in Object(arr)) {
      return sliceIterator(arr, i);
    } else {
      throw new TypeError("Invalid attempt to destructure non-iterable instance");
    }
  };
}();













var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

var ncurry = function ncurry(n) {
  var localCurry = function localCurry(f) {
    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    return function () {
      for (var _len2 = arguments.length, largs = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        largs[_key2] = arguments[_key2];
      }

      var rest = args.concat(largs);

      if (rest.length < n) return localCurry.apply(undefined, [f].concat(toConsumableArray(rest)));
      return f.apply(undefined, toConsumableArray(rest));
    };
  };
  return localCurry;
};

var curry2 = ncurry(2);
var curry3 = ncurry(3);
var curry4 = ncurry(4);

var clone = function clone(obj) {
  var hash = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : new WeakMap();

  if (Object(obj) !== obj) return obj; // primitives
  if (hash.has(obj)) return hash.get(obj); // cyclic references
  var result = void 0;

  if (Array.isArray(obj)) result = [];else if (obj.constructor) result = new obj.constructor();else result = Object.create(null);

  hash.set(obj, result);

  if (obj instanceof Map) Array.from(obj, function (_ref) {
    var _ref2 = slicedToArray(_ref, 2),
        key = _ref2[0],
        val = _ref2[1];

    return result.set(key, clone(val, hash));
  });
  return Object.assign.apply(Object, [result].concat(toConsumableArray(Object.keys(obj).map(function (key) {
    return defineProperty({}, key, clone(obj[key], hash));
  }))));
};

var promisify = function promisify(f) {
  return function () {
    for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
      args[_key] = arguments[_key];
    }

    return Promise.all(args.map(future)).then(function (params) {
      return f.apply(undefined, toConsumableArray(params));
    });
  };
};

/**
 * Run a function for side effect and return the original value. The original
 * value gets cloned and therefore can be modified in the tap handler.
 *
 * @example
 * const f = a => future(a);
 * flow([f, tap(console.log)])(23); // Print "23" to the console.
 *
 * @param {(Promise.<Function>|Function)} f The function for side effect. The
 * return value of this function will be ignored.
 * @param {(Promise.<*>|*)} x The value to use for side effect. This can
 * either be a value or a promise that resolves to a value.
 * @returns {Promise.<*>} Returns `x`.
 */
var tap = curry2(promisify(function (f, x) {
  return future(f(clone(x))).then(function () {
    return x;
  });
}));

/**
 * Map a function over a promise.
 *
 * @param {(Promise.<Function>|Function.<*>)} f The function to apply. `f` can
 * either be a function, or a promise that resolves to a promise. The function
 * can either return a value, or a promise that resolves to a promise.
 * @param {(Promise.<*>|*)} x The value to apply to `f`. This can either be a
 * value, or a promise that resolves to a value.
 * @returns {Promise.<*>} The result of `x` applied to `f`.
 */
var fmap = curry2(promisify(function (f, x) {
  return future(x).then(f);
}));

/**
 * Apply a function wrapped in a promise to a promisified value.
 *
 * @param {Promise.<Function>} pf A promise that resolves to a function.
 * @param {Promise.<Any>} p A promise that resolves to a value.
 * @returns {Promise.<Any>} A promise resolving to x applied to the function
 * that f resolves to.
 * @example
 * const pf = future(v => v + 1);
 * const p = future(1);
 * apply(pf, f).then(console.log); // Returns 2.
 */
var apply = curry2(promisify(function (pf, p) {
  return fmap(function (f) {
    return fmap(f, p);
  }, pf);
}));

/**
 * Create a function that always returns the same value.
 *
 * @example
 * const f = constant("Hello");
 * f().then(console.log); // Returns "Hello"
 *
 * @param {(Promise.<*>|*)} x The value to return. This can either be a value
 * or a promise for a value.
 * @returns {Promise.<*>} A promise that resolves to `x`.
 */
var constant = function constant(x) {
  return function () {
    return future(x);
  };
};

/**
 * Create a function that evaluates all promises in an array when called. This
 * is equivalent to `Promise.all`, with the difference that it creates a
 * callable function.
 *
 * @example
 * const f = all([openFile1(), opeFile2(), openFile3()]);
 * f().then(console.log); // Returns [a, b, c];
 *
 * @param {(Promise.<Array>|Array.<*>)} xs An array of values that are
 * resolved. `xs` can either be an array, or a promise that resolves to an
 * array. Each element of `xs` can either be a value, or a promise that
 * resolves to a value.
 * @returns {Function.<Promise.Array<*>>} A function, that when called, resolves all
 * promises and returns an Array of values.
 */
var all = function all(xs) {
  return function () {
    return future(xs).then(function (ys) {
      return Promise.all(ys.map(future));
    });
  };
};

/**
 * Compose two function that return promises to yield a third function that
 * returns a promise. The resulting composite function is denoted
 * `g∘f : X → Z`, defined by `(g∘f)(x) = g(f(x))` for all `x` in `X`.
 *
 * @example
 * const f = x => future(x + 1);
 * const g = x => future(x + 5);
 * const h = compose(f, g);
 * h(10).then(console.log); // 10 + 1 + 5, returns 16.
 *
 * @param {(Promise.<Function>|Function)} f The first function to
 * compose. This argument can either be a function, or a promise that resolves
 * to a function. The function can either return a value or a promise for a
 * value.
 * @param {(Promise.<Function>|Function)} g The second function to
 * compose. This argument can either be a function, or a promise that resolves
 * to a function. The function can either return a value or a promise for a
 * value.
 * @param {(Promise|*)} x The argument to call `(g∘f)` with. It can either be
 * any value, or the promise for any value.
 * @returns {Promise.<*>} The result of calling `g` with the result of
 * `f(x)`.
 */
var compose = curry3(promisify(function (f, g, x) {
  return future(f(x)).then(g);
}));

/**
 * Create a function out of a list of functions, where each successive
 * invocation is supplied the return value of the previous function call. The
 * new function forms a pipe where the results flow from left to right so to
 * speak. It's a shortcut for composing more than two functions.
 *
 * @example
 * const f = (x, y) => future(x + y);
 * const fs = map(f, [...Array(5).keys()]);
 * flow(fs, 0).then(console.log); // The sum of [0, 1, 2, 3, 4], returns 10
 *
 * @param {(Promise.<Array>|Array.<Function>)} fs An array of functions to
 * compose. `fs` can either be an array, or a promise that resolves to an
 * array. Each function can either return a value or a promise for a value.
 * @param {(Promise|*)} x The argument to call the function pipeline with. It
 * can either be any value, or the promise for any value.
 * @returns {Promise.<*>} The result of applying `x` to the pipeline of
 * `fs`.
 */
var flow = curry2(promisify(function (fs, x) {
  return fs.reduce(function (memo, f) {
    return memo.then(f);
  }, future(x));
}));

/**
 * Reduce a list of values to a single value, using a reduction function. This
 * is equivalent to `Array.reduce`.
 *
 * @example
 * const f = (acc, x) => future(acc + x);
 * const xs = [...Array(5).keys()];
 * fold(f, 0, xs).then(console.log); // The sum of [0, 1, 2, 3, 4], returns 10;
 *
 * @param {(Promise.<Function>|Function.<*>)} f The reduce function to map
 * over the list. `f` can either be a function, or a promise that resolves to
 * a promise. This function can either return a value or the promise for a
 * value.
 * @param {(Promise.<*>|*)} acc The initial value to apply to the first call
 * of `f`. It can either be any value or the promise for any value.
 * @param {(Promise.<Array>|Array.<*>)} xs The list to reduce to a single
 * value. This can either be an array, or the promise for an array.
 * @returns {Promise.<*>} The value of `xs` reduced over `f`.
 */
var fold = curry3(promisify(function (f, acc, xs) {
  return xs.reduce(function (memo, x) {
    return memo.then(function (y) {
      return f(y, x);
    });
  }, future(acc));
}));

var mapN = curry2(promisify(function (concurrency, f, xs) {
  return (
    // eslint-disable-next-line promise/avoid-new
    new Promise(function (resolve, reject) {
      var ret = [];
      var iterator = xs[Symbol.iterator]();
      var isRejected = false;
      var iterableDone = false;
      var resolvingCount = 0;
      var currentIdx = 0;

      var next = function next() {
        if (isRejected) return;
        var nextItem = iterator.next();
        var i = currentIdx;
        currentIdx += 1;

        if (nextItem.done) {
          iterableDone = true;

          if (resolvingCount === 0) resolve(ret);
          return;
        }
        resolvingCount += 1;

        Promise.resolve(nextItem.value).then(f)
        // eslint-disable-next-line promise/always-return
        .then(function (val) {
          ret[i] = val;
          resolvingCount -= 1;
          next(); // eslint-disable-line promise/no-callback-in-promise
        }).catch(function (err) {
          isRejected = true;
          reject(err);
        });
      };

      for (var i = 0; i < concurrency; i += 1) {
        next();
        if (iterableDone) break;
      }
    })
  );
}));

/**
 * Map a function over every element of a list. This is equivalent to
 * `Array.map`. Only one promise at a time get's resolved.
 *
 * @example
 * const f = x => future(x + 1);
 * const xs = [...Array(5).keys()];
 * map(f, xs).then(console.log); // Prints [1, 2, 3, 4, 5]
 *
 * @param {Function.<*>} f The function that is applied to every element. This
 * function can either return a value or the promise for a value.
 * @param {(Promise.<Array>|Array.<*>)} xs The list to map `f` over. This can
 * either be an array, or the promise for an array.
 * @returns {Promise.<Array>} A list of the same length as `xs`, but with `f`
 * applied to each of its elements.
 */
var map = mapN(1);

/**
 * The same as `map`, but run two promises concurrently.
 */
var map2 = mapN(2);

/**
 * The same as `map`, but run three promises concurrently.
 */
var map3 = mapN(3);

/**
 * The same as `map`, but run four promises concurrently.
 */
var map4 = mapN(4);

/**
 * The same as `map`, but run five promises concurrently.
 */
var map5 = mapN(5);

// Flattens an array of arrays into an array. It only flattens one level deep,
// which is what is required by flatmap.
var flatten = (function (xxs) {
  var result = [];

  if (xxs == null) return result;

  xxs.forEach(function (xs) {
    if (Array.isArray(xs)) result.push.apply(result, toConsumableArray(xs));else result.push(xs);
  });

  return result;
});

var flatmapN = curry3(function (mapper, f, xs) {
  return mapper(f, xs).then(flatten);
});

/**
 * Map a function over every element of a list and concatenate the results
 * into a single list. Only one promise at a time get's resolved.
 *
 * @example
 * const f = x => [x, x];
 * const xs = [1, 2];
 * flatmap(f, xs).then(console.log); // Prints [1, 1, 2, 2];
 *
 * @param {Function.<*>} f The function that is applied to every element. This
 * function can either return a value or the promise for a value.
 * @param {(Promise.<Array>|Array.<*>)} xs The list to map `f` over. This can
 * either be an array, or the promise for an array.
 * @returns {Promise.<Array>} The concatenation of applying every element of
 * `xs` to `f`.
 */
var flatmap = flatmapN(map);

/**
 * The same as `flatmap`, but run two promises concurrently.
 */
var flatmap2 = flatmapN(map2);

/**
 * The same as `flatmap`, but run three promises concurrently.
 */
var flatmap3 = flatmapN(map3);

/**
 * The same as `flatmap`, but run four promises concurrently.
 */
var flatmap4 = flatmapN(map4);

/**
 * The same as `flatmap`, but run five promises concurrently.
 */
var flatmap5 = flatmapN(map5);

/**
 * Catch an exception on a promise and call a exception handler. This is
 * equivalent to `Promise.catch`.
 *
 * @example
 * const f = () => new Error("Boom");
 * caught(console.error, f()); // Prints the exception.
 *
 * @param {Function} f The handler to call when the promise `p` get's
 * rejected.
 * @param {Promise.<*>|*} p A value or a promise that get's resolved.
 * @returns {Promise.<*>} A promise that resolves either to the value of `p`
 * or to the return value of `f`.
 */
var caught = curry2(function (f, p) {
  return future(p).catch(f);
});

var identity = (function (x) {
  return x;
});

/**
 * Make a conditional test and either call the left or the right branch.
 *
 * @example
 * const pred = userExists;
 * const cons = updateUser;
 * const alt = createUser;
 * whenElse(userExists, updateUser, createUser, user);
 *
 * @param {Function} predicate Test if a condition is `true` or `false`.
 * @param {Function} consequent Apply `value` to this function if the
 * predicate return `true`;
 * @param {Function} alternative Apply `value` to this function if the
 * predicate return `false`;
 * @param {Promise|*} value Test this value to decide whether to call the
 * consequent or the alternative.
 * @returns {Promise.<*>} A promise that resolves to a value, that is the
 * result of either calling the `consequent` or the `alternative` with
 * `value`.
 */
var whenElse = curry4(promisify(function (predicate, consequent, alternative, value) {
  var chooseFn = function chooseFn(bool) {
    return bool === true ? consequent(value) : alternative(value);
  };
  return future(predicate(value)).then(chooseFn);
}));

/**
 * Like `whenElse`, only call the `consequent` if the predicate returns
 * `false`, and the `alternative` if the predicate returns `true`;
 */
var unlessElse = curry4(function (predicate, consequent, alternative, value) {
  return whenElse(predicate, alternative, consequent, value);
});

/**
 * Like `whenElse`, but have no alternative. If the predicate returns `false`,
 * simply return the `value`.
 */
var when = curry3(function (predicate, consequent, value) {
  return whenElse(predicate, consequent, identity, value);
});

/**
 * Like `unlessElse`, but have no alternative. If the predicate returns `true`,
 * simply return the `value`.
 */
var unless = curry3(function (predicate, consequent, value) {
  return unlessElse(predicate, consequent, identity, value);
});

/**
 * Lift a binary function over two promises.
 *
 * @param {(Promise.<Function>|Function.<*, *>)} f A binary function. This
 * can either be a function, or a promise that resolves to a function. The
 * function can either return a value, or a promise that resolves to a value.
 * @param {(Promise.<*>|*)} x A value that gets lifted as the first argument
 * of `f`. This can either be a value, or a promise that resolves to a value.
 * @param {(Promise.<*>|*)} y A value that gets lifted as the first argument
 * of `f`. This can either be a value, or a promise that resolves to a value.
 * @returns {Promise.<*>} The value that f returns when applied to `x` and
 * `y`.
 * @example
 * const f = (x, y) => future(x = y);
 * const a = future(1);
 * const b = future(2);
 * lift2(f, a, b).then(console.log); // Returns 3.
 */
var lift2 = curry3(promisify(function (f, x, y) {
  return fmap(fmap(function (u) {
    return function (v) {
      return f(u, v);
    };
  }, x), y);
}));

/**
 * Call a variadic function with the value of a promise as it's arguments. If
 * the value is an array, flatten it to the formal parameters of the
 * fulfillment handler.
 *
 * @example
 * const add = (x, y) => x + y;
 * const p = future([1, 2]);
 * spread(add, p).then(console.log); // Prints 3
 *
 * @param {(Promise.<Function>|Function.<*>)} f The function to apply to the
 * value of p. This function can either return a value or the promise for a
 * value. This function can also be a promise that resolves to a function.
 * @param {(Promise.<*>|*)} p The promise that resolves to the arguments for
 * the function. This can either be a single value, an Array or a promise
 * resolving to any of those.
 * @returns {Promise.<*>} The result of applying the value of `p` to `f`.
 */
var spread = curry2(promisify(function (f, p) {
  return future(p).then(function (args) {
    if (Array.isArray(args)) return f.apply(undefined, toConsumableArray(args));
    return f(args);
  });
}));

/**
 * Delay the resolution of a promise chain.
 *
 * @example
 * delay(100).then(console.log) // Waits 100 ms.
 *
 * @param {number} ms The time to wait in milliseconds.
 * @param {Promise.<*>|*} value The promise or value to resolve after the
 * delay.
 * @returns {Promise} A promise that resolved to value, or whatever value
 * resolves to.
 */
var delay = curry2(function (ms, value) {
  var cancel = void 0;

  // eslint-disable-next-line promise/avoid-new
  var thunk = new Promise(function (resolve, reject) {
    var timeOut = setTimeout(function () {
      return resolve(value);
    }, ms);
    cancel = function cancel() {
      if (timeOut) {
        clearTimeout(timeOut);
        timeOut = null;
      }
      reject(new Error("Promise canceled"));
    };
  });
  thunk.cancel = cancel;
  return thunk;
});

/**
 * Test if an object is a promise.
 *
 * @example
 * const p = future(23);
 * isPromise(p); // Returns true;
 *
 * @param {*} p The object to test.
 * @returns {boolean} Returns `true` if the object is a promise, otherwise
 * `false`;
 */
var isPromise = function isPromise(p) {
  return !!p && ((typeof p === "undefined" ? "undefined" : _typeof(p)) === "object" || typeof p === "function") && typeof p.then === "function";
};

// times is 0-indexed. A times value of 4, actually makes 5 retries.
var retrier = function retrier(times, waitTime) {
  return function (action) {
    for (var _len = arguments.length, args = Array(_len > 1 ? _len - 1 : 0), _key = 1; _key < _len; _key++) {
      args[_key - 1] = arguments[_key];
    }

    var delayModifier = 1.5;
    var f = isPromise(action) ? function () {
      return action;
    } : action;

    var resolver = function resolver(iter, waitMs) {
      if (iter <= 0) return f.apply(undefined, args);
      return caught(function () {
        return delay(waitMs, null).then(function () {
          return resolver(iter - 1, waitMs * delayModifier);
        });
      }, f.apply(undefined, args));
    };

    return resolver(times, waitTime);
  };
};

/**
 * Call an action, and retry it in case it fails. An action is retried up to
 * five times with an increasing timeout.
 *
 * @example
 * // Retries `fetchUser` in case of failure.
 * retry(fetchUser).then(console.log).catch(console.error);
 *                        
 *
 * @param {Promise|Function} action A function or a promise that should be
 * retried in case of failure.
 * @returns {Promise.<*>} A promise that resolves to the return value of
 * `action`, or that is rejected with the last exception that `action` failed
 * with.
 */
var retry = retrier(4, 250);

// Using curry allows to fix the number of arguments for functions that are
// implemented as variadic functions.
/**
 * Like `retry`, but accept one additional argument that is applied to
 * `action`.
 *
 * @example
 * // Retries `fetchUser` in case of failure.
 * retry(fetchUser, 23).then(console.log).catch(console.error);
 *
 * @param {Promise|Function} action A function or a promise that should be
 * retried in case of failure.
 * @param {*} arg The argument to apply to `action`.
 * @returns {Promise.<*>} A promise that resolves to the return value of
 * `action`, or that is rejected with the last exception that `action` failed
 * with.
 */
var retry2 = curry2(retrier(4, 250));

/**
 * Like `retry2`, but accept two arguments to apply to `action`.
 */
var retry3 = curry3(retrier(4, 250));
/**
 * Like `retry2`, but accept three arguments to apply to `action`.
 */
var retry4 = curry4(retrier(4, 250));

exports.future = future;
exports.tap = tap;
exports.fmap = fmap;
exports.apply = apply;
exports.constant = constant;
exports.all = all;
exports.compose = compose;
exports.flow = flow;
exports.fold = fold;
exports.map = map;
exports.map2 = map2;
exports.map3 = map3;
exports.map4 = map4;
exports.map5 = map5;
exports.flatmap = flatmap;
exports.flatmap2 = flatmap2;
exports.flatmap3 = flatmap3;
exports.flatmap4 = flatmap4;
exports.flatmap5 = flatmap5;
exports.caught = caught;
exports.when = when;
exports.whenElse = whenElse;
exports.unless = unless;
exports.unlessElse = unlessElse;
exports.lift2 = lift2;
exports.spread = spread;
exports.delay = delay;
exports.retry = retry;
exports.retry2 = retry2;
exports.retry3 = retry3;
exports.retry4 = retry4;
exports.isPromise = isPromise;

Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=combinatorsP.js.map
